import pickle
import numpy
import chessboard
import park_martin
import yaml
numpy.set_printoptions(linewidth=300, suppress=True)
from scipy.linalg import expm, inv
from numpy import dot, eye
import sys

if sys.version_info.major > 2:
    img_list = pickle.load(open('data/image_list.dump', 'rb'), encoding = 'latin1')
    rob_pose_list = pickle.load(open('data/pose_list.dump', 'rb'), encoding = 'latin1')
else:
    img_list = pickle.load(open('data/image_list.dump', 'rb'))
    rob_pose_list = pickle.load(open('data/pose_list.dump', 'rb'))

corner_list = []
obj_pose_list = []

camera_matrix, dist_coeffs = chessboard.calibrate_lens(img_list)

def hat(v):
    return [[   0, -v[2],  v[1]],
            [v[2],     0, -v[0]],
            [-v[1],  v[0],    0]]

def tf_mat(r, t):
    res = eye(4)
    res[0:3, 0:3] = expm(hat(r))
    res[0:3, -1] = t
    return res

for i, img in enumerate(img_list):
    found, corners = chessboard.find_corners(img)
    corner_list.append(corners)
    if not found:
        raise Exception("Failed to find corners in img # %d" % i)
    rvec, tvec = chessboard.get_object_pose(chessboard.pattern_points, corners, camera_matrix, dist_coeffs)
    object_pose = tf_mat(rvec, tvec)
    obj_pose_list.append(object_pose)

A, B = [], []
for i in range(1,len(img_list)):
    p = rob_pose_list[i-1], obj_pose_list[i-1]
    n = rob_pose_list[i], obj_pose_list[i]
    A.append(dot(inv(p[0]), n[0]))
    B.append(dot(inv(p[1]), n[1]))


# Transformation to chessboard in robot gripper
X = eye(4)
Rx, tx = park_martin.calibrate(A, B)
X[0:3, 0:3] = Rx
X[0:3, -1] = tx

print("X: ")
print(X)

print("For validation. Printing transformations from the robot base to the camera")
print("All the transformations should be quite similar")

for i in range(len(img_list)):
    rob = rob_pose_list[i]
    obj = obj_pose_list[i]
    tmp = dot(rob, dot(X, inv(obj)))
    print(tmp)

# Here I just pick one, but maybe some average can be used instead
rob = rob_pose_list[0]
obj = obj_pose_list[0]
cam_pose = dot(dot(rob, X), inv(obj))

cam = {'rotation' : cam_pose[0:3, 0:3].tolist(),
       'translation' : cam_pose[0:3, -1].tolist(),
       'camera_matrix' : camera_matrix.tolist(),
       'dist_coeffs' : dist_coeffs.tolist()}

fp = open('camera.yaml', 'w')
fp.write(yaml.dump(cam))
fp.close()



import numpy as np
import ipywidgets

from tpk4170.visualization.viewer import Viewer
from tpk4170.forward_kinematics import fk_kr6r900sixx
from tpk4170.models import kr6r900sixx as kr6
from tpk4170.models import Grid, Ball


class Visualizer:
    def __init__(self, base_link, link_1, link_2,
                 link_3, link_4, link_5, link_6, fk, interact=False):
        self.viewer = Viewer()
        self.grid = Grid()
        self.viewer.add(self.grid)

        self.base_link = base_link
        self.link_1 = link_1
        self.link_2 = link_2
        self.link_3 = link_3
        self.link_4 = link_4
        self.link_5 = link_5
        self.link_6 = link_6

        self.viewer.add(self.base_link)
        self.viewer.add(self.link_1)
        self.viewer.add(self.link_2)
        self.viewer.add(self.link_3)
        self.viewer.add(self.link_4)
        self.viewer.add(self.link_5)
        self.viewer.add(self.link_6)

        self._fk = fk

        if interact:
            self.interact()

        self.show(np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]))

    def show(self, q, show_trajectory=False):
        T01, T02, T03, T04, T05, T06 = self._fk(q)
        self.link_1(T01)
        self.link_2(T02)
        self.link_3(T03)
        self.link_4(T04)
        self.link_5(T05)
        self.link_6(T06)

        if show_trajectory:
            ball = Ball(color='white', radius=0.01)
            ball.position = T06[:3, 3].tolist()
            self.viewer.add(ball)

    def interact(self):
        def f(q1, q2, q3, q4, q5, q6):
            q = np.array([q1, q2, q3, q4, q5, q6])
            self.show(q)

        ipywidgets.interact(f,
                            q1=(np.deg2rad(-180), np.deg2rad(180)),
                            q2=(np.deg2rad(-180), np.deg2rad(180)),
                            q3=(np.deg2rad(-180), np.deg2rad(180)),
                            q4=(np.deg2rad(-180), np.deg2rad(180)),
                            q5=(np.deg2rad(-180), np.deg2rad(180)),
                            q6=(np.deg2rad(-180), np.deg2rad(180)))


class Kr6R900SixxVisualizer(Visualizer):
    def __init__(self, interact=False, show_trajectory=False):
        Visualizer.__init__(self,
                            kr6.BaseLink(),
                            kr6.Link1(),
                            kr6.Link2(),
                            kr6.Link3(),
                            kr6.Link4(),
                            kr6.Link5(),
                            kr6.Link6(),
                            fk_kr6r900sixx,
                            interact)




# Author: Torstein A. Myhre

import numpy
import cv2

pattern_size = (10, 6)
square_size = 0.02

pattern_points = numpy.zeros((numpy.prod(pattern_size), 3), numpy.float32)
pattern_points[:, :2] = numpy.indices(pattern_size).T.reshape(-1, 2)
pattern_points *= square_size


def find_corners(image):
    found, corners = cv2.findChessboardCorners(image, pattern_size)
    term = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_COUNT, 30, 0.1)
    if found:
        cv2.cornerSubPix(image, corners, (5, 5), (-1, -1), term)
    return found, corners


def draw_corners(image, corners):
    color_image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
    cv2.drawChessboardCorners(color_image, pattern_size, corners, True)
    return color_image


def get_object_pose(object_points, image_points, camera_matrix, dist_coeffs):
    ret, rvec, tvec = cv2.solvePnP(
        object_points, image_points, camera_matrix, dist_coeffs)
    return rvec.flatten(), tvec.flatten()


def calibrate_lens(image_list):
    img_points, obj_points = [], []
    h, w = 0, 0
    for img in image_list:
        h, w = img.shape[:2]
        found, corners = find_corners(img)
        if not found:
            raise Exception(
                "chessboard calibrate_lens Failed to find corners in img")
        img_points.append(corners.reshape(-1, 2))
        obj_points.append(pattern_points)
    camera_matrix = numpy.zeros((3, 3))
    dist_coeffs = numpy.zeros(5)
    cv2.calibrateCamera(obj_points, img_points, (w, h),
                        camera_matrix, dist_coeffs)
    return camera_matrix, dist_coeffs



import cv2 as cv
import numpy as np
from numpy import sqrt, dot, cross, vstack
from numpy.linalg import norm
from tpk4170.utils.transformations import unit_vector

# Implementaton of trilaterate taken from:
# https://stackoverflow.com/questions/1406375/finding-intersection-points-between-3-spheres


def trilaterate(P1, P2, P3, r1, r2, r3):
    '''
    Find the intersection of three spheres

    P1,P2,P3 are the centers, r1,r2,r3 are the radii
    '''
    temp1 = P2-P1
    e_x = temp1/norm(temp1)
    temp2 = P3-P1
    i = dot(e_x, temp2)
    temp3 = temp2 - i*e_x
    e_y = temp3/norm(temp3)
    e_z = cross(e_x, e_y)
    d = norm(P2-P1)
    j = dot(e_y, temp2)
    x = (r1*r1 - r2*r2 + d*d) / (2*d)
    y = (r1*r1 - r3*r3 - 2*i*x + i*i + j*j) / (2*j)
    temp4 = r1*r1 - x*x - y*y
    if temp4 < 0:
        raise Exception("The three spheres do not intersect!")
    z = sqrt(temp4)
    p_12_a = P1 + x*e_x + y*e_y + z*e_z
    p_12_b = P1 + x*e_x + y*e_y - z*e_z
    return vstack((p_12_a, p_12_b))


def solveP3P(obj_points, img_points, camera_matrix, dist_coeffs):
    '''
    Implementation of the solution to the P3P problem as presented in 
    Fischler and Bolles, Random Sample Consensus: A Paradigm for Model
    Fitting with Applications to Image Analysis and Automated Cartography.

    The function uses three object and image correspondences to compute a
    set of solutions. A fourth point is used as a control point. The first
    solution found is returned.
    '''

    obj_points = obj_points.reshape(-1, 3)

    img_points = cv.undistortPoints(img_points, camera_matrix, dist_coeffs)
    img_points = img_points.reshape(-1, 2)

    op = obj_points[:3].reshape(3, 3)
    ip = img_points[:3].reshape(3, 2)
    hip = np.hstack((ip, np.ones((3, 1))))

    # Compute cosines
    p, q, r = hip
    Cab = np.inner(unit_vector(p), unit_vector(q))
    Cac = np.inner(unit_vector(p), unit_vector(r))
    Cbc = np.inner(unit_vector(q), unit_vector(r))

    # Compute distances
    A, B, C = op
    Rab = np.linalg.norm(B-A)
    Rac = np.linalg.norm(C-A)
    Rbc = np.linalg.norm(C-B)

    # Compute coefficients
    K1 = Rbc**2 / Rac**2
    K2 = Rbc**2 / Rab**2
    G4 = (K1*K2-K1-K2)**2 - 4*K1*K2*Cbc**2
    G3 = 4*(K1*K2-K1-K2)*K2*(1-K1)*Cab + 4*K1 * \
        Cbc*((K1*K2-K1+K2)*Cac + 2*K2*Cab*Cbc)
    G2 = (2*K2*(1-K1)*Cab)**2 + 2*(K1*K2-K1-K2)*(K1*K2+K1-K2) + 4*K1 * \
        ((K1-K2)*Cbc**2 + K1*(1-K2)*Cac**2 - 2*(1+K1)*K2*Cab*Cac*Cbc)
    G1 = 4*(K1*K2+K1-K2)*K2*(1-K1)*Cab + 4*K1 * \
        ((K1*K2-K1+K2)*Cac*Cbc + 2*K1*K2*Cab*Cac**2)
    G0 = (K1*K2+K1-K2)**2 - 4*K1**2*K2*Cac**2

    # Solve for x
    xs = np.roots((G4, G3, G2, G1, G0))
    xs = np.real(xs[xs.imag < 1e-5])

    # Solve for lengths a, b, and c
    # Compute a and b
    as_ = Rab / np.sqrt(xs**2 - 2*xs*Cab + 1)
    bs = as_ * xs

    def y(x, a, b):
        m = 1-K1
        p = 2*(K1*Cac-x*Cbc)
        q = x**2-K1
        m_ = 1
        p_ = 2*(-x*Cbc)
        q_ = (x**2*(1-K2)+2*x*K2*Cab-K2)
        num = (p_*q - p*q_)
        den = (m*q_ - m_*q)
        return num / den

    # Compute c
    cs = y(xs, as_, bs) * as_

    # Concatenate a,b,c
    lengths = np.dstack((as_, bs, cs)).reshape(-1, 3)

    # Solve for T (the perspective center)
    Ts = []
    for (a, b, c) in lengths:
        Ts.append(trilaterate(A, B, C, a, b, c))
    Ts = np.array(Ts).reshape(-1, 3)

    # Compute lambdas
    lambdas = []
    hip_norm = np.linalg.norm(hip, axis=1)
    for T in Ts:
        lambdas.append(np.linalg.norm(op - T, axis=1) / hip_norm)
    lambdas = np.array(lambdas).reshape(-1, 3)

    # Compute R (rotation matrix)
    Rs = []
    for lambda_, T in zip(lambdas, Ts):
        V = hip.T * lambda_.reshape(3, 1)
        X = op.T - T.reshape(3, 1)
        R = V @ np.linalg.inv(X)
        Rs.append(R)

    # Return first correct solution
    rvecs = []
    tvecs = []
    cop = obj_points[3]
    cip = img_points[3]
    for R, t in zip(Rs, Ts):
        # x, y, z = (R @ cop.reshape(3, 1) + t.reshape(3, 1)).ravel()
        # if np.allclose(cip, (x/z, y/z)):
        rvecs.append(cv.Rodrigues(R)[0])
        tvecs.append(t.reshape(3, 1))

    return len(rvecs), rvecs, tvecs



import numpy as numpy


def project(obj_points, rvec, tvec, camera_matrix, dist_coeffs):
    obj = np.vstack((obj_points.T, np.ones(len(obj_points))))
    T = np.eye(4)
    T[:3, :3] = cv.Rodrigues(rvec)[0]
    T[:3, 3] = tvec.ravel()
    P = np.eye(3, 4)
    K = camera_matrix
    hom = (K @ P @ T @ obj).T
    return np.array([(u/w, v/w) for u, v, w in hom])


def hom(p):
    '''Homogenous form of p'''
    return np.vstack((p.T, np.ones(len(p))))


def distort(img_points, camera_matrix, dist_coeffs):
    '''Apply distortion to image points'''
    k1, k2, p1, p2, k3 = dist_coeffs.ravel()
    ones = np.ones(len(img_points))
    p = np.vstack((img_points.T, ones))
    x = np.linalg.inv(camera_matrix) @ p
    u, v = x[:2, :]
    r = np.sqrt(u**2 + v**2)
    delta_u = u * (k1*r**2 + k2*r**4 + k3*r**6) + \
        2.*p1*u*v + p2*(r**2 + 2.*u**2)
    delta_v = v * (k1*r**2 + k2*r**4 + k3*r**6) + \
        p1*(r**2 + 2.*v**2) + 2*p2*u*v
    ud = u + delta_u
    vd = v + delta_v
    xd = np.vstack((ud, vd, ones))
    pd = camera_matrix @ xd
    return pd.T[:, :2]



import numpy as np
from tpk4170.forward_kinematics import Jacobian


def ik_analytical_ur5(trf, shoulder=-1, elbow=1, wrist=-1):

    a = [0.00000, -0.42500, -0.39225,  0.00000,  0.00000,  0.0000]
    d = [0.089159,  0.00000,  0.00000,  0.10915,  0.09465,  0.0823]

    Te = trf
    ne, se, ae, pe = Te[:3, :].T

    d1, _, _, d4, d5, d6 = d
    _, a2, a3, _, _, _ = a

    L2 = a2
    L3 = a3

    p5 = pe - d6 * ae
    p5x, p5y, _ = p5

    psi5 = np.arctan2(shoulder * p5y, shoulder * p5x)
    offset1 = np.arctan2(d4, np.sqrt(p5x*p5x + p5y*p5y - d4*d4))

    q1 = psi5 + shoulder * offset1

    x1 = np.array([np.cos(q1), np.sin(q1), 0])
    y1 = np.array([0, 0, 1])
    z1 = np.array([np.sin(q1), -np.cos(q1), 0])

    z4 = wrist * shoulder * np.cross(z1, ae)
    z4 /= np.linalg.norm(z4)
    y4 = z1
    x4 = np.cross(y4, z4)

    q5 = np.arctan2(np.inner(-ae, x4), np.inner(ae, y4))

    q6 = np.arctan2(np.inner(-z4, ne), np.inner(-z4, se))

    p3x, p3y, p3z = p5 - d5 * z4 - d4 * z1

    pv = p3z - d1
    ph = shoulder * np.sqrt(p3x*p3x + p3y*p3y)

    q234 = np.arctan2(np.inner(z4, x1), np.inner(z4, -y1))

    c3 = (ph*ph + pv*pv - L2*L2 - L3*L3) / (2 * L2 * L3)
    s3 = elbow * np.sqrt(1.0 - c3*c3)
    q3 = np.arctan2(s3, c3)

    c2 = (ph * (L2 + L3*c3) + pv*L3*s3) / (L2*L2 + L3*L3 + 2*L2*L3*c3)
    s2 = (pv * (L2 + L3*c3) - ph*L3*s3) / (L2*L2 + L3*L3 + 2*L2*L3*c3)
    q2 = np.arctan2(s2, c2)

    q4 = q234 - (q2 + q3)
    if q4 > np.pi:
        q4 -= 2 * np.pi

    return np.array([q1, q2, q3, q4, q5, q6])


def ik_iterative(trf, fk, q0=np.array([1, 1, 1, 1, 1, 1])):
    Rd = trf[:3, :3]
    qk = q0
    qs = []
    for i in range(1000):
        Ts = fk(qk)
        Tk = Ts[5]
        Jk = Jacobian(Ts)
        Rk = Tk[:3, :3]
        Re = Rd @ Rk.T
        ep = trf[:3, 3] - Tk[:3, 3]
        eo = np.array([Re[2, 1] - Re[1, 2],
                       Re[0, 2] - Re[2, 0],
                       Re[1, 0] - Re[0, 1]]) * 0.5

        e = np.array((ep, eo)).reshape(6, 1)

        K = 0.1
        dq = K * np.dot(np.linalg.pinv(Jk), e).ravel()
        qk = qk + dq
        qs.append(qk)

        if np.linalg.norm(e) < 1e-6:
            break
    return qk



import sys
from pythreejs import Object3D, AxesHelper
import numpy as np

from tpk4170.models import ColladaMesh
from tpk4170.utils.transformations import quaternion_from_matrix


class KR6R900sixx:
    def __init__(self):
        pass


class Link(Object3D):
    def __init__(self):
        Object3D.__init__(self)

    def __call__(self, trf):
        self.quaternion = quaternion_from_matrix(trf).tolist()
        self.position = (trf[:3, 3]).tolist()


dae_path = sys.prefix + '/share/tpk4170/models/kr6r900sixx/visual/'


class BaseLink(Link):
    def __init__(self):
        Object3D.__init__(self)
        mesh = ColladaMesh(dae_path + 'base_link.dae', scale=0.001)
        self.add(mesh)
        axes = AxesHelper(size=0.5)
        self.add(axes)


class Link1(Link):
    def __init__(self):
        Link.__init__(self)
        mesh = ColladaMesh(dae_path + 'link_1.dae', scale=0.001)
        mesh.rotateX(np.pi/2)
        mesh.position = (-0.025, 0.0, 0.0)
        self.mesh = mesh
        self.add(mesh)
        axes = AxesHelper(size=0.5)
        self.add(axes)


class Link2(Link):
    def __init__(self):
        Link.__init__(self)
        mesh = ColladaMesh(dae_path + 'link_2.dae', scale=0.001)
        mesh.rotateZ(np.pi/2)
        mesh.rotateX(np.pi/2)
        mesh.position = (-0.455, 0.0, 0.0)
        self.mesh = mesh
        self.add(mesh)
        axes = AxesHelper(size=0.5)
        self.add(axes)


class Link3(Link):
    def __init__(self):
        Link.__init__(self)
        mesh = ColladaMesh(dae_path + 'link_3.dae', scale=0.001)
        mesh.position = (-0.035, 0.0, -0.025)
        mesh.rotateZ(np.pi)
        mesh.rotateY(-np.pi/2)
        self.mesh = mesh
        self.add(mesh)


class Link4(Link):
    def __init__(self):
        Link.__init__(self)
        mesh = ColladaMesh(dae_path + 'link_4.dae', scale=0.001)
        mesh.rotateZ(np.pi/2)
        mesh.rotateX(np.pi/2)
        mesh.position = (0.0, -0.42, 0)
        self.mesh = mesh
        self.add(mesh)
        axes = AxesHelper(size=0.5)
        self.add(axes)


class Link5(Link):
    def __init__(self):
        Link.__init__(self)
        mesh = ColladaMesh(dae_path + 'link_5.dae', scale=0.001)
        mesh.rotateX(np.pi)
        mesh.rotateY(np.pi/2)
        self.mesh = mesh
        self.add(mesh)
        axes = AxesHelper(size=0.5)
        self.add(axes)


class Link6(Link):
    def __init__(self):
        Link.__init__(self)
        mesh = ColladaMesh(dae_path + 'link_6.dae', scale=0.001)
        mesh.rotateY(-np.pi/2)
        self.mesh = mesh
        self.add(mesh)
        axes = AxesHelper(size=0.5)
        self.add(axes)


class Joint:
    pass


class Chain:
    pass



from . files import *


def process_image(fn, debug=False):

    #  Read image from file
    img = cv.imread(fn, 0)
    if img is None:
        print("Failed to load", fn)
        return None

    #  Locate chessboard corners in images
    found, corners = cv.findChessboardCorners(img, pattern_size)
    if found:
        term = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_COUNT, 30, 0.001)
        cv.cornerSubPix(img, corners, (5, 5), (-1, -1), term)

    #  Debug: Draw chessboard on image
    if debug:
        vis = cv.cvtColor(img, cv.COLOR_GRAY2BGR)
        cv.drawChessboardCorners(vis, pattern_size, corners, found)
        _path, name, _ext = splitfn(fn)
        outfile = os.path.join(_path, name + '_chess.png')
        cv.imwrite(outfile, vis)

    #  Return None if the chessboard is not founf
    if not found:
        print('Chessboard not found')
        return None

    #  Print status
    print('{}... OK'.format(fn))

    return (corners.reshape(-1, 2), pattern_points)



import numpy as np
import pythreejs
from pythreejs import Group
from pythreejs import Mesh
from pythreejs import (Geometry, SphereGeometry,
                       BufferGeometry, PlaneGeometry)
from pythreejs import BufferAttribute
from pythreejs import Points, PointsMaterial
from pythreejs import (
    LineBasicMaterial, MeshLambertMaterial, MeshPhongMaterial, MeshBasicMaterial)
from pythreejs import AxesHelper as Axes

from matplotlib.colors import to_hex
from collada import Collada


class Grid(Group):
    def __init__(self, num_cells=5, color='#cccccc', linewidth=1, cellsize=0.5):
        Group.__init__(self)
        material = LineBasicMaterial(color=color, linewidth=linewidth)
        for i in range(num_cells+1):
            edge = cellsize * num_cells / 2
            position = edge - (i * cellsize)
            geometry_h = Geometry(vertices=[(-edge, position, 0),
                                            (edge, position, 0)])
            geometry_v = Geometry(vertices=[(position, -edge, 0),
                                            (position, edge, 0)])
            self.add(pythreejs.Line(geometry=geometry_h, material=material))
            self.add(pythreejs.Line(geometry=geometry_v, material=material))


class Ball(Mesh):
    def __init__(self, color='red', radius=0.025):
        Mesh.__init__(self, geometry=SphereGeometry(radius=radius),
                      material=MeshLambertMaterial(color=color))


class ColladaMesh(Group):
    def __init__(self, filename, scale=1.0):
        Group.__init__(self)

        self._dae = Collada(filename)
        self._load_mesh(self._dae, scale=scale)

    def _load_mesh(self, dae, scale):
        materials = self._load_material(dae)
        for geometry in dae.geometries:
            for primitive in geometry.primitives:
                vertices = primitive.vertex[primitive.vertex_index] * scale
                normals = primitive.normal[primitive.normal_index]
                buffer_geometry = BufferGeometry(
                    attributes={'position': BufferAttribute(array=vertices),
                                'normal': BufferAttribute(array=normals)})
                material = materials[primitive.material]
                mesh = Mesh(geometry=buffer_geometry, material=material)
                self.add(mesh)

    def _load_material(self, dae):
        materials = {}
        for material in dae.materials:
            name = material.id
            color = to_hex(material.effect.diffuse)
            specular = to_hex(material.effect.specular)
            materials[name] = MeshPhongMaterial(color=color,
                                                specular=specular,
                                                shininess=30)
        return materials


class Plane(Mesh):
    def __init__(self, color='pink', transparent=False):
        Mesh.__init__(self, geometry=PlaneGeometry(),
                      material=MeshBasicMaterial(color=color))
        self.material.side = 'DoubleSide'
        if transparent:
            self.material.transparent = transparent
            self.material.opacity = 0.5


class Line(pythreejs.Line):
    def __init__(self, start, end, color='white', linewidth=1):
        geometry = BufferGeometry(attributes={
            'position': BufferAttribute(
                np.vstack((start, end)).astype(np.float32),
                normalized=False)
        })
        material = LineBasicMaterial(color=color, linewidth=linewidth)
        pythreejs.Line.__init__(self, geometry=geometry, material=material)


class Triangle(Mesh):
    def __init__(self, p1, p2, p3, color='yellow'):
        geometry = BufferGeometry(
            attributes={
                'position': BufferAttribute(
                    np.vstack((p1, p2, p3)).reshape(3, 3).astype(np.float32),
                    normalized=False)
            }
        )
        material = MeshBasicMaterial(color=color)
        material.side = 'DoubleSide'
        Mesh.__init__(self, geometry=geometry, material=material)


class PointCloud(Points):
    def __init__(self, points, point_size=0.001, color='green'):
        geometry = BufferGeometry(
            attributes={'position': BufferAttribute(array=points.astype(np.float32))})
        material = PointsMaterial(size=point_size, color=color)
        Points.__init__(self, geometry=geometry, material=material)



import numpy as np
import matplotlib.pyplot as plt


def plot_trajectory(trajectory, title=''):
    qs, dqs, ddqs, ts = trajectory
    f, axarr = plt.subplots(3, sharex=True)
    axarr[0].plot(ts, qs, color='red')
    axarr[0].set_title(title)
    axarr[0].set_ylabel('position')
    axarr[1].plot(ts, dqs, color='green')
    axarr[1].set_ylabel('velocity')
    axarr[2].plot(ts, ddqs, color='blue')
    axarr[2].set_ylabel('acceleration')
    axarr[2].set_xlabel('time [s]')


def quintic_trajectory(current_position, target_position,
                       current_velocity, target_velocity,
                       current_acceleration, target_acceleration,
                       duration_in_seconds):
    trajectories = []
    t = duration_in_seconds
    xs = np.linspace(0, t)
    for qi, qf, dqi, dqf, ddqi, ddqf in zip(current_position, target_position,
                                            current_velocity, target_velocity,
                                            current_acceleration, target_acceleration):
        A = np.array(
            [[0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
             [t**5, t**4, t**3, t**2, t, 1.0],
             [0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
             [5. * t**4, 4. * t**3, 3. * t**2, 2. * t, 1., 0.0],
             [0.0, 0.0, 0.0, 2.0, 0.0, 0.0],
             [20. * t**3, 12. * t**2, 6. * t, 2., 0.0, 0.0]])

        b = np.array([qi, qf, dqi, dqf, ddqi, ddqf])
        x = np.linalg.solve(A, b)

        qs = np.polyval(x, xs)
        dqs = np.polyval(
            [5. * x[0], 4. * x[1], 3. * x[2], 2. * x[3], x[4]], xs)
        ddqs = np.polyval([20. * x[0], 12. * x[1], 6. * x[2], 2. * x[3]], xs)

        trajectories.append((qs, dqs, ddqs, xs))
    return trajectories


def cubic_trajectory(current_position, target_position,
                     current_velocity, target_velocity,
                     duration_in_seconds):
    trajectories = []
    t = duration_in_seconds
    ts = np.linspace(0, t)
    for qi, qf, dqi, dqf in zip(current_position, target_position,
                                current_velocity, target_velocity):
        A = np.array([[0.0, 0.0, 0.0, 1.0],
                      [t**3, t**2, t, 1.],
                      [0.0, 0.0, 1.0, 0.0],
                      [3.0 * t**2, 2*t, 1.0, 0.0]])

        b = np.array([qi, qf, dqi, dqf])
        x = np.linalg.solve(A, b)

        qs = np.polyval(x, ts)
        dqs = np.polyval([3. * x[0], 2. * x[1], x[2]], ts)
        ddqs = np.polyval([6. * x[0], 2. * x[1]], ts)

        trajectories.append((qs, dqs, ddqs, ts))
    return trajectories


class LSPBError(Exception):
    pass


def lspb_trajectory(current_position, target_position,
                    duration_in_seconds, cruise_velocity=None):

    q0 = current_position
    q1 = target_position
    tf = duration_in_seconds
    V = cruise_velocity

    ts = np.linspace(0.0, tf)

    if V is None:
        V = (q1-q0)/tf * 1.5
    else:
        V = np.abs(V) * np.sign(q1-q0)
        if np.abs(V) < np.abs(q1-q0)/tf:
            raise LSPBError('V too small')
        elif np.abs(V) > 2 * np.abs(q1-q0)/tf:
            raise LSPBError('V too big')

    if np.allclose(q0, q1):
        s = np.ones(len(ts)) * q0
        sd = np.zeros(len(ts))
        sdd = np.zeros(len(ts))
        return (s, sd, sdd)

    tb = (q0 - q1 + V*tf)/V
    a = V/tb

    p = []
    pd = []
    pdd = []

    for tt in ts:
        if tt <= tb:
            # Initial blend
            p.append(q0 + a/2*tt*tt)
            pd.append(a*tt)
            pdd.append(a)
        elif tt <= (tf - tb):
            p.append((q1+q0-V*tf)/2 + V*tt)
            pd.append(V)
            pdd.append(0.0)
        else:
            p.append(q1 - (a/2*tf*tf) + a*tf*tt - a/2*tt*tt)
            pd.append(a*tf - a*tt)
            pdd.append(-a)

    return (p, pd, pdd, ts)



import numpy as np

I = np.eye(3)


def vec(a, b, c):
    return np.array([[a], [b], [c]])


def skew(v):
    return np.array([[0, -v[2], v[1]],
                     [v[2], 0, -v[0]],
                     [-v[1], v[0], 0]])


def vee(S):
    return vec([S[2, 1], S[0, 2], S[1, 0]])


def exp(v):
    angle = np.linalg.norm(v)
    S = skew(v / angle)
    return I + np.sin(angle) * S + (1 - np.cos(angle)) * S @ S



from pythreejs import (DirectionalLight, PointLight, AmbientLight,
                       PerspectiveCamera, Scene, OrbitControls, Renderer)
from IPython.display import display


class Viewer:
    def __init__(self, width=768, height=512, near=0.01, far=1000, background="#111111", antialias=True):

        key_light = DirectionalLight(color='white', position=[
                                     3, 3, 3], intensity=0.66)

        c = PerspectiveCamera(40, width/height, near, far)
        c.position = [3, 3, 3]
        c.up = [0, 0, 1]

        c.add(key_light)

        pl = PointLight(color='white', intensity=0.1, position=[3, 3, 3])

        self._scene = Scene()
        self._scene.background = background
        self._scene.add(AmbientLight())
        self._scene.add(pl)

        renderer = Renderer(camera=c,
                            scene=self._scene,
                            antialias=antialias,
                            controls=[OrbitControls(controlling=c)],
                            height=height, width=width)
        display(renderer)

    def add(self, obj):
        self._scene.add(obj)



#!/usr/bin/env python
# Copyright (c) NTNU, Norwegian University of Science and Technology
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(name='tpk4170',
      version='0.0.2',
      description='Python modules for the course TPK4170 Robotics at NTNU ',
      author='Lars Tingelstad',
      author_email='lars.tingelstad@ntnu.no',
      url='',
      download_url='https://github.com/tingelst/tpk4170-robotics/archive/master.zip',
      packages=find_packages(),
      install_requires=requirements,
      data_files=[
          ('share/tpk4170/models/ur5/visual/',
             [
                 'tpk4170/models/ur5/visual/base.dae',
                 'tpk4170/models/ur5/visual/forearm.dae',
                 'tpk4170/models/ur5/visual/shoulder.dae',
                 'tpk4170/models/ur5/visual/upperarm.dae',
                 'tpk4170/models/ur5/visual/wrist1.dae',
                 'tpk4170/models/ur5/visual/wrist2.dae',
                 'tpk4170/models/ur5/visual/wrist3.dae',
             ]),
          ('share/tpk4170/models/kr6r900sixx/visual/',
           [
               'tpk4170/models/kr6r900sixx/visual/base_link.dae',
               'tpk4170/models/kr6r900sixx/visual/link_1.dae',
               'tpk4170/models/kr6r900sixx/visual/link_2.dae',
               'tpk4170/models/kr6r900sixx/visual/link_3.dae',
               'tpk4170/models/kr6r900sixx/visual/link_4.dae',
               'tpk4170/models/kr6r900sixx/visual/link_5.dae',
               'tpk4170/models/kr6r900sixx/visual/link_6.dae',
           ]),
          ('share/tpk4170/models/kr16_2/visual/',
           [
               'tpk4170/models/kr16_2/visual/base_link.dae',
               'tpk4170/models/kr16_2/visual/link_1.dae',
               'tpk4170/models/kr16_2/visual/link_2.dae',
               'tpk4170/models/kr16_2/visual/link_3.dae',
               'tpk4170/models/kr16_2/visual/link_4.dae',
               'tpk4170/models/kr16_2/visual/link_5.dae',
               'tpk4170/models/kr16_2/visual/link_6.dae',
           ]),
      ],
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      )



import numpy as np
from tpk4170.models import Grid, Ball, Axes
from tpk4170.visualization import Viewer
from tpk4170.models.ur5 import BaseLink, Link1, Link2, Link3, Link4, Link5, Link6
import ipywidgets


def dh(a, alpha, d, theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    ca = np.cos(alpha)
    sa = np.sin(alpha)
    return np.array([[ct, -st * ca,  st * sa, a * ct],
                     [st,  ct * ca, -ct * sa, a * st],
                     [0.0,      sa,       ca, d],
                     [0.0,     0.0,      0.0, 1.0]])


def fk_dh(dh_a, dh_alpha, dh_d, q_zero_offset, q):
    A1 = dh(dh_a[0], dh_alpha[0], dh_d[0], q[0] + q_zero_offset[0])
    A2 = dh(dh_a[1], dh_alpha[1], dh_d[1], q[1] + q_zero_offset[1])
    A3 = dh(dh_a[2], dh_alpha[2], dh_d[2], q[2] + q_zero_offset[2])
    A4 = dh(dh_a[3], dh_alpha[3], dh_d[3], q[3] + q_zero_offset[3])
    A5 = dh(dh_a[4], dh_alpha[4], dh_d[4], q[4] + q_zero_offset[4])
    A6 = dh(dh_a[5], dh_alpha[5], dh_d[5], q[5] + q_zero_offset[5])
    T01 = A1
    T02 = A1 @ A2
    T03 = A1 @ A2 @ A3
    T04 = A1 @ A2 @ A3 @ A4
    T05 = A1 @ A2 @ A3 @ A4 @ A5
    T06 = A1 @ A2 @ A3 @ A4 @ A5 @ A6
    return (T01, T02, T03, T04, T05, T06)


def Jacobian(Ts):
    p0 = np.array([0, 0, 0])
    z0 = np.array([0, 0, 1])
    pn = Ts[5][:3, 3]
    J = np.zeros((6, 6))
    J[:3, 0] = np.cross(z0, pn-p0)
    J[3:, 0] = z0
    for i in range(5):
        T = Ts[i]
        zi = T[:3, 2]
        pi = T[:3, 3]
        J[:3, i+1] = np.cross(zi, pn-pi)
        J[3:, i+1] = zi
    return J

def fk(q):
    a = [0.00000, -0.42500, -0.39225,  0.00000,  0.00000,  0.0000]
    d = [0.089159,  0.00000,  0.00000,  0.10915,  0.09465,  0.0823]
    alpha = [ 1.570796327, 0, 0, 1.570796327, -1.570796327, 0 ]
    q_zero_offset = [0, 0, 0, 0, 0, 0]
    return fk_dh(a, alpha, d, q_zero_offset, q)

class UR5Visualizer:
    def __init__(self, interact=False):
        self.viewer = Viewer()
        self.grid = Grid()
        self.viewer.add(self.grid)
        
        self.base_link = BaseLink()
        self.viewer.add(self.base_link)
        self.link_1 = Link1()
        self.viewer.add(self.link_1)
        self.link_2 = Link2()
        self.viewer.add(self.link_2)
        self.link_3 = Link3()
        self.viewer.add(self.link_3)
        self.link_4 = Link4()
        self.viewer.add(self.link_4)
        self.link_5 = Link5()
        self.viewer.add(self.link_5)
        self.link_6 = Link6()
        self.viewer.add(self.link_6)
        
        if interact:
            self.interact()
            
        self.show(np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]))
        
    def show(self, q):
        T01, T02, T03, T04, T05, T06 = fk(q)
        self.link_1(T01)
        self.link_2(T02)
        self.link_3(T03)
        self.link_4(T04)
        self.link_5(T05)
        self.link_6(T06)
        
        
    def interact(self):
        def f(q1, q2, q3, q4, q5, q6):
            q = np.array([q1, q2, q3, q4, q5, q6])
            self.show(q)
            
        ipywidgets.interact(f,
                            q1=(np.deg2rad(-180), np.deg2rad(180)),
                            q2=(np.deg2rad(-180), np.deg2rad(180)),
                            q3=(np.deg2rad(-180), np.deg2rad(180)),
                            q4=(np.deg2rad(-180), np.deg2rad(180)),
                            q5=(np.deg2rad(-180), np.deg2rad(180)),
                            q6=(np.deg2rad(-180), np.deg2rad(180)))
        
        



# Implements the method described in:
# @article{park_robot_1994,
#  title = {Robot sensor calibration: solving {AX}= {XB} on the {Euclidean} group},
#  volume = {10},
#  number = {5},
#  journal = {IEEE Transactions on Robotics and Automation},
#  author = {Park, Frank C and Martin, Bryan J},
#  year = {1994},
#  pages = {717--721}
# }

# Author: Torstein A. Myhre

import numpy
from numpy import dot, eye, zeros, outer
from numpy.linalg import inv


def log(R):
    # Rotation matrix logarithm
    theta = numpy.arccos((R[0, 0] + R[1, 1] + R[2, 2] - 1.0)/2.0)
    return numpy.array([R[2, 1] - R[1, 2],
                        R[0, 2] - R[2, 0],
                        R[1, 0] - R[0, 1]]) * theta / (2*numpy.sin(theta))


def invsqrt(mat):
    u, s, v = numpy.linalg.svd(mat)
    return u.dot(numpy.diag(1.0/numpy.sqrt(s))).dot(v)


def calibrate(A, B):
    # transform pairs A_i, B_i
    N = len(A)
    M = numpy.zeros((3, 3))
    for i in range(N):
        Ra, Rb = A[i][0:3, 0:3], B[i][0:3, 0:3]
        M += outer(log(Rb), log(Ra))

    Rx = dot(invsqrt(dot(M.T, M)), M.T)

    C = zeros((3*N, 3))
    d = zeros((3*N, 1))
    for i in range(N):
        Ra, ta = A[i][0:3, 0:3], A[i][0:3, 3]
        Rb, tb = B[i][0:3, 0:3], B[i][0:3, 3]
        C[3*i:3*i+3, :] = eye(3) - Ra
        d[3*i:3*i+3, 0] = ta - dot(Rx, tb)

    tx = dot(inv(dot(C.T, C)), dot(C.T, d))
    return Rx, tx.flatten()



import os
from glob import glob
import numpy as np
import cv2 as cv


def load_images(mask='./data/images/image??.png'):
    '''Load images in sorted order based on filename'''
    img_names = glob(mask)
    img_names.sort(key=lambda s: int(os.path.basename(s)[5:-4]))
    img_list = [cv.imread(img_name, 0) for img_name in img_names]
    return img_list


def load_rob_poses(mask='./data/poses/pose??.txt'):
    '''Load poses in sorted order based on filename'''
    pose_names = glob(mask)
    pose_names.sort(key=lambda s: int(os.path.basename(s)[4:-4]))
    rob_pose_list = [np.loadtxt(pose_name) for pose_name in pose_names]
    return rob_pose_list
